from typing import Any, TypeVar, Union, Type

import pkg_resources

from kikyo.settings import Settings

T = TypeVar('T')


class Kikyo:
    def __init__(self, _settings: dict = None, **kwargs):
        self.settings = Settings(_settings)
        self.settings.merge(kwargs)
        self._components = {}
        self._init()

    def _init(self):
        self._init_plugins()

    def _init_plugins(self):
        plugins = {
            entry_point.name: entry_point.load()
            for entry_point in pkg_resources.iter_entry_points('kikyo.plugins')
        }

        active_plugins = self.settings.getlist('active_plugins')
        if active_plugins:
            active_plugins = set(active_plugins)
            for name in list(plugins.keys()):
                if name not in active_plugins:
                    del plugins[name]

        for name, plugin in plugins.items():
            if hasattr(plugin, 'configure_kikyo'):
                plugin.configure_kikyo(self)

    def add_component(self, name: str, component: Any):
        self._components[name] = component

    def component(self, name: str = None, cls: Type[T] = None) -> Union[T, Any]:
        if name is not None:
            if name not in self._components:
                raise RuntimeError(f"No component named '{name}'")
            return self._components[name]
        candidates = []
        for v in self._components.values():
            if isinstance(v, cls):
                candidates.append(v)
        if len(candidates) == 0:
            raise RuntimeError(f"No such type component: {cls}")
        if len(candidates) > 1:
            raise RuntimeError(f"Got {len(candidates)} candidates of {cls}")
        return candidates[0]



from abc import ABCMeta, abstractmethod
from typing import Any


class Bucket(metaclass=ABCMeta):

    @abstractmethod
    def get_object_link(self, key: str) -> str:
        """
        获取对象的下载链接

        :param key: 对象的名称
        """

    @abstractmethod
    def put_object(self, key: str, data: Any):
        """
        上传对象

        :param key: 对象的名称
        :param data: 对象数据
        """


class OSS(metaclass=ABCMeta):
    """
    提供对象存储服务
    """

    @abstractmethod
    def bucket(self, name: str) -> Bucket:
        """
        获取bucket

        :param name: bucket名称
        """



from abc import ABCMeta, abstractmethod
from typing import Any


class Producer(metaclass=ABCMeta):
    """
    生产者
    """

    def __enter__(self) -> 'Producer':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abstractmethod
    def send(self, *records: Any):
        """
        发送数据

        :param records: 数据
        """

    @abstractmethod
    def close(self):
        """
        关闭生产者
        """


class Message(metaclass=ABCMeta):
    """
    消息
    """

    @property
    @abstractmethod
    def value(self) -> Any:
        """
        消息内容
        """


class Consumer(metaclass=ABCMeta):
    """
    消费者
    """

    def __enter__(self) -> 'Consumer':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abstractmethod
    def receive(self) -> Message:
        """
        接收数据
        """

    @abstractmethod
    def close(self):
        """
        关闭消费者
        """

    @abstractmethod
    def ack(self, msg: Message):
        """
        确认消息

        :param msg: 消息
        """


class DataHub(metaclass=ABCMeta):
    """
    提供数据总线服务
    """

    @abstractmethod
    def create_producer(self, topic: str) -> Producer:
        """
        创建向指定topic发送数据的生产者

        :param topic: topic名称
        """

    @abstractmethod
    def subscribe(
            self,
            topic: str,
            subscription_name: str = None,
            auto_ack: bool = True,
    ) -> Consumer:
        """
        订阅指定topic

        :param topic: topic名称
        :param subscription_name: 订阅的标识
        :param auto_ack: 自动确认
        """
